//<head>//////////////////////////////////////////////////////////////
//
//	
//	Lexed version 4
//	Copyright (C) 1999 2000 2001 2002 2003.
//	Universit Paris 7
//	INRIA Rocquencourt
//
//	Conception & programmation :
//	Lionel Clment
//	e-mail : lionel.clement@linguist.jussieu.fr
//
//
////////////////////////////////////////////////////////////////</head>////

#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "lexedtree.h"
#include "lexedinfo.h"

int LexedMemory;
extern int LexedIndexation;
extern char *Sep_Pref;
extern char *Sep_Or;
extern char *Sep_Suff;
extern char *Sep_Uw;
extern char *LexedBuffer;
extern char Port[255];
extern char Mot[LEXEDMAXSTRING];
extern int swap;
extern TYPEPTR LexedInitial;
extern Fsa      *fsa;    // table fsa
extern InfoBuff *info;   // table des info
extern TYPEPTR OffsetFsa[8];
extern TYPEPTR OffsetInfo[2];
FILE *Table;

static off_t nbreItem;
static off_t nbreItemMax=0;
static struct stat *statbuf;
static char Delimiter[LEXEDMAXSTRING];
static int NbreApproxErase=0;
static int NbreApproxSubstitute=0;
static int NbreApproxAdd=0;
static int NbreApproxSwap=0;
static int NbreApproxEraseDouble=0;
static int NbreApproxAddDouble=0;
static int NbreApproxReplace=0;
TYPEPTR MaxWeight=(~0UL);

static struct FindReplaceStructure *Replace;
int ReplaceIndex;
int ReplaceSize;

extern Arbre *Lexique;
extern Arbre *LexiqueInit;
  
////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////
void Usage() {
  fprintf(stderr,
    "Usage: lexed [global-option] <build|consult> [mode-options] <filename>\n"
    "Global options:\n"
    "-h|--help                              print this\n"
    "-v|--version                           print version\n"
    "-p <name>                              prefix name\n"
    "-d <directory>                         directory\n"
    "Options for build mode:\n"
    "-i                                     indexation mode\n"
    "Options for consult mode:\n"
    "-f <pref> <or> <suff> <uw>             format information\n"
    "-P <port>                              port number (daemon)\n"
    "-F                                     look for information in a file rather than in memory\n"
    "-t <chars>                             delimiter characters\n"
    "-e <number>                            number of characters to erase\n"
    "-s <number>                            number of characters to subtitute\n"
    "-a <number>                            number of characters to add\n"
    "-w <number>                            number of characters to swap\n"
    "-b <number>                            number of double characters to erase\n"
    "-B <number>                            number of double characters to add\n"
    "-S <number>                            number of characters to translate\n"
    "-r <chars to seach> <chars to replace> add a charactere translation schema\n"
    "-W <number>                            the maximum weight\n"
  );
  exit (0);
}

////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////
int min (int a, int b){
  if (a>b) return b; else return a;
}

int Build(int argn, char **argv, int pos, char* Directory, char* Prefix) {

  FILE *Input;
  FILE *FSA;
  char InputFile[LEXEDMAXSTRING];
  char TableFile[LEXEDMAXSTRING];
  char FSAFile[LEXEDMAXSTRING];
  char chaine [LEXEDMAXSTRING];
  char graphie [LEXEDMAXSTRING];
  char oldChaine [LEXEDMAXSTRING];
  int i;
  unsigned int rang;
  TYPEPTR offset;
  
  offset = 0;
  rang = 0;

  // mode specific options
  for (i = pos; argv[i]; i++) {
    if (argv[i][0] == '-') {
      if (!strcmp(argv[i]+1, "i")) {
	LexedIndexation = 1;
	strcpy (Sep_Uw, "0");
      }
    } else {
      strcpy (InputFile, argv[i]);
    }
  }

  if (!LexedIndexation) {
    sprintf (TableFile, "%s/%s.tbl", Directory, Prefix);
    Table = fopen (TableFile, "w");
    if (!Table) {
      fprintf (stderr, "Unable to open file %s for writing\n", TableFile);
      exit (0);
    }
  }

  if (InputFile && strcmp(InputFile, "")) {
    fprintf(stderr, "*** Loading lexicon %s\n", InputFile);
    Input = fopen (InputFile, "r");
    if (!Input) {
      fprintf (stderr, "Unable to open file %s for reading\n", InputFile);
      exit (0);
    }
  } else {
    fprintf(stderr, "*** Loading lexicon from stdin\n");
    Input = stdin;
  }

  statbuf=(struct stat *)malloc(sizeof(struct stat));
  stat(InputFile, statbuf);
  nbreItemMax=statbuf->st_size;
  free(statbuf);

  if (Lexique == NULL) {
    Lexique=new Arbre(NULL, NULL, NULL, '\0');
    Lexique->fils=new Arbre(NULL, NULL, NULL, '\0');
    LexiqueInit=Lexique;
    Lexique=Lexique->fils;
  }

  while (fgets(chaine, LEXEDMAXSTRING, Input)) {
    if (!*chaine) {
      continue;
    }

    for (i = 0; i <= strlen(chaine) && (!strchr (Delimiter, chaine[i])); i++)
      /* empty */;

    chaine[strlen(chaine)-1] = 0;
    strcpy (graphie, chaine);
    if (i < strlen (chaine)) {
      graphie[i]=0;
      if (LexedIndexation) {
	Lexique->Add(graphie, ++offset);
      } else {
	if (!strcmp(chaine+i+1, oldChaine)) {
	  Lexique->Add(graphie, offset);
	} else {
	  // Adds one entry
	  offset = ftell(Table);
	  Lexique->Add(graphie, offset);
	  fprintf(Table, "%s%c", chaine+i+1, 0);
	}
	strcpy (oldChaine, chaine+i+1); 
      }
    } else {
      // the entry of one word without any information is the same as the previous:
      // aaa info_of_aaa
      // bbb
      //
      // bbb -> info_of_aaa
      if (LexedIndexation) {
	Lexique->Add(graphie, ++offset);
      } else {
	Lexique->Add(graphie, offset);
      }
    }

    nbreItem+=strlen(chaine)+1;
    if (!(rang++ % 1357)) {
      fprintf (stderr, " %.0f%%\r", (100*(float)nbreItem)/(float)nbreItemMax);
    }
  }

  if (InputFile) {
    fclose (Input);
  }

  if (!LexedIndexation) {
    fclose (Table);
  }

  sprintf (FSAFile, "%s/%s.fsa", Directory, Prefix);
  //cerr << "*** type: " << sizeof(TYPEPTR)*8 << "\n";
  fputs("*** Writing Finite State Automata\n", stderr);
  FSA = fopen (FSAFile, "w");
  if (FSA) {
    SaveFSA(FSA);
    fclose (FSA);
  } else {
    fprintf (stderr, "Unable to open file %s for writing\n", FSAFile);
    exit (0);
  }
}

int Consult(int argn, char **argv, int pos, char* Directory, char* Prefix) {

  FILE *Input;
  FILE *Table;
  FILE *FSA;
  char InputFile[LEXEDMAXSTRING];
  char TableFile[LEXEDMAXSTRING];
  char FSAFile[LEXEDMAXSTRING];
  TcpServer *myserver;
  char chaine [LEXEDMAXSTRING];
  int i;
  int lists;
  char *buf;
  int ok;
  unsigned long int Taille;
  TYPEPTR info;

  lists = 0;

  // mode specific options
  for (i = pos; argv[i]; i++) {
    if (argv[i][0] == '-') {
	if (!strcmp(argv[i]+1, "f")) {	
	  if (i+4 >= argn || argv[i+1][0] == '-' || argv[i+2][0] == '-' || argv[i+3][0] == '-' || argv[i+4][0] == '-') {
	    Usage();
	  } else {
	    strcpy (Sep_Pref, argv[++i]);
	    strcpy (Sep_Or, argv[++i]);
	    strcpy (Sep_Suff, argv[++i]);
	    strcpy (Sep_Uw, argv[++i]);
	  }
	} else if (!strcmp(argv[i]+1, "P")) {
	  if ((i+1 >= argn) || (argv[i+1][0]=='-')) {
	    Usage();
	  } else {
	    strcpy (Port, argv[++i]);
	  }
	} else if (!strcmp(argv[i]+1, "F")) {
	  LexedMemory = 0;
	} else if (!strcmp(argv[i]+1, "t")) {
	  if (i+1 >= argn || argv[i+1][0] == '-') {
	    Usage();
	  } else {
	    strcpy (Delimiter, argv[++i]);
	  }
	} else if (!strcmp(argv[i]+1, "e")) {
	  if (i+1 >= argn || argv[i+1][0] == '-' || atoi(argv[i+1]) == 0) {
	    Usage();
	  } else {
	    NbreApproxErase=atoi(argv[++i]);
	  }
	} else if (!strcmp(argv[i]+1, "s")) {
	  if ((i+1 >= argn) || (argv[i+1][0] == '-') || (atoi(argv[i+1]) == 0)) {
	    Usage();
	  } else {
	    NbreApproxSubstitute=atoi(argv[++i]);
	  }
	} else if (!strcmp(argv[i]+1, "S")) {
	  if (i+1 >= argn || argv[i+1][0] == '-' || atoi(argv[i+1]) == 0) {
	    Usage();
	  } else {
	    NbreApproxReplace=atoi(argv[++i]);
	  }
	} else if (!strcmp(argv[i]+1, "a")) {
	  if (i+1 >= argn || argv[i+1][0] == '-' || atoi(argv[i+1]) == 0) {
	    Usage();
	  } else {
	    NbreApproxAdd=atoi(argv[++i]);
	  }
	} else if (!strcmp(argv[i]+1, "w")) {
	  if (i+1 >= argn || argv[i+1][0] == '-' || atoi(argv[i+1]) ==0 ) {
	    Usage();
	  } else {
	    NbreApproxSwap=atoi(argv[++i]);
	  }
	} else if (!strcmp(argv[i]+1, "b")) {
	  if (i+1 >= argn || argv[i+1][0] == '-' || atoi(argv[i+1]) == 0) {
	    Usage();
	  } else {
	    NbreApproxEraseDouble=atoi(argv[++i]);
	  }
	} else if (!strcmp(argv[i]+1, "B")) {
	  if (i+1 >= argn || argv[i+1][0] == '-' || atoi(argv[i+1]) == 0) {
	    Usage();
	  } else {
	    NbreApproxAddDouble=atoi(argv[++i]);
	  }
	} else if (!strcmp(argv[i]+1, "r")) {
	  if (i+2 >= argn || argv[i+1][0] == '-' || argv[i+2][0] == '-') {
	    Usage();
	  } else {
	    if (Replace == NULL) {
	      ReplaceSize=LEXEDMAXREPLACE;
	      Replace = (struct FindReplaceStructure *)calloc(sizeof(struct FindReplaceStructure), LEXEDMAXREPLACE);
	      ReplaceIndex=0;
	      Replace[ReplaceIndex].Find=NULL;
	      Replace[ReplaceIndex].Repl=NULL;
	    }
	    if (ReplaceIndex == ReplaceSize) {
	      ReplaceSize *= 2;
	      Replace = (struct FindReplaceStructure *)realloc(Replace, sizeof(struct FindReplaceStructure) * ReplaceSize);
	      Replace[ReplaceIndex].Find=NULL;
	      Replace[ReplaceIndex].Repl=NULL;
	    }
	    Replace[ReplaceIndex].Find=(char *)calloc(strlen(argv[++i])+1, sizeof(char));
	    strcpy(Replace[ReplaceIndex].Find, argv[i]);
	    Replace[ReplaceIndex].Repl=(char *)calloc(strlen(argv[++i])+1, sizeof(char));
	    strcpy(Replace[ReplaceIndex].Repl, argv[i]);
	    ReplaceIndex++;
	  }
	} else if (!strcmp(argv[i]+1, "W")) {
	  if (i+1 >= argn || argv[i+1][0] == '-' || atoi(argv[i+1]) == 0) {
	    Usage();
	  } else {
	    MaxWeight=(TYPEPTR)atoi(argv[++i]);
	  }
	} else if (!strcmp(argv[i]+1, "l")) {
	  lists = 1;
	}
      } else {
      strcpy (InputFile, argv[i]);
    }
  }

  if (Port[0]){
    myserver = (TcpServer*)malloc(sizeof(TcpServer));
    *myserver = TcpServer(Port);
  }

  sprintf (FSAFile, "%s/%s.fsa", Directory, Prefix);
  FSA = fopen(FSAFile, "r");
  if (!FSA) {
    fprintf(stderr, "File %s not found\n", FSAFile);
    exit (0);
  } else {
    if (!LoadFSA(FSA)) {
      return 1;
    } 
    fclose (FSA);
  }

  sprintf (TableFile, "%s/%s.tbl", Directory, Prefix);
  Table = fopen (TableFile, "r");
  if (!Table) {
    fprintf(stderr, "File %s not found\n", TableFile);
    exit (0);
  } else {
    if (LexedMemory) {
      fputs("*** Load buffer in memory\n", stderr);
      statbuf = (struct stat *)malloc(sizeof(struct stat));
      stat(TableFile, statbuf);
      Taille = statbuf->st_size;
      free(statbuf);
      
      LexedBuffer = new char [Taille];
      LexedBuffer[0] = 0;
      
      ok = fread (LexedBuffer, 1, Taille, Table);
      fclose (Table);
    } else {
      fputs("*** Load buffer as a file\n", stderr);
    }
  }

  if (lists) {
    char *Chaine = (char *)malloc (5*1024*1024);
    //strcpy(Chaine, "                     ");
    strcpy(Chaine, "");
    List(fsa[LexedInitial].fils, 0, myserver, Chaine, 0, Table);
    strcpy(Chaine, "");
    free(Chaine);
    return 0;
  }

  if (Port[0]) {
    fprintf(stderr, "*** Ready for connection with a client on port %s\n", Port);
#ifdef DAEMON
    daemon(0,0);
#endif
    myserver->Start();
    while (myserver->GetData()!=-1) {
      buf=myserver->GetBuffer();
      if (buf[0]==-1)
	break;
      //strcpy(chaine,buf);
      if (NbreApproxErase || NbreApproxSubstitute || NbreApproxAdd || NbreApproxSwap || NbreApproxEraseDouble || NbreApproxAddDouble || NbreApproxReplace){
	info = SearchStatic(LexedInitial, chaine);
	if (info!=(TYPEPTR)~0UL)
	  PrintResults(info, 1, myserver, 1, Table);
	else {
	  int out=0;
	  int ResultSize=0;
	  int ResultIndex=0;
	  TYPEPTR *Result=NULL;
	  Search(LexedInitial, 1, myserver, buf, NbreApproxErase, NbreApproxSubstitute, NbreApproxAdd, NbreApproxSwap, NbreApproxEraseDouble, NbreApproxAddDouble, NbreApproxReplace, Replace, 0, &Result, &ResultSize, &ResultIndex, 1);
	  for (TYPEPTR *R=Result; R!=NULL && *R!=(TYPEPTR)~0UL; R++, R++){
	    if ((MaxWeight == (~0UL)) ||
		((*(R+1)) <= MaxWeight)) {
	      if (out) {
		myserver->PutData(Sep_Or);
	      }
	      out = 1;
	      PrintResults ((*R), 1, myserver, 0, Table);
	    }
	  }
	  if (!out) {
	    myserver->PutData(Sep_Uw);
	  }
	  myserver->PutData(Sep_Suff);
	}
      } else {
	info = SearchStatic(LexedInitial, buf);
	PrintResults(info, 1, myserver, 1, Table);
      }
    }
    myserver->Stop();
  } else {
    if (InputFile && strcmp(InputFile, "")) {
      fprintf(stderr, "*** Seaching %s\n", InputFile);
      Input = fopen (InputFile, "r");
      if (!Input) {
	fprintf (stderr, "Unable to open file %s for reading\n", InputFile);
	exit (0);
      }
    } else {
      fprintf(stderr, "*** Searching from stdin\n");
      Input = stdin;
    }
    while (fgets (chaine, LEXEDMAXSTRING, Input)) {
      //
      chaine[strlen(chaine)-1]=0;
      if (NbreApproxErase || NbreApproxSubstitute || NbreApproxAdd || NbreApproxSwap || NbreApproxEraseDouble || NbreApproxAddDouble || NbreApproxReplace){
	int out=0;
	int ResultSize=0;
	int ResultIndex=0;
	TYPEPTR *Result=NULL;
	Search(LexedInitial, 0, myserver, chaine, NbreApproxErase, NbreApproxSubstitute, NbreApproxAdd, NbreApproxSwap, NbreApproxEraseDouble, NbreApproxAddDouble, NbreApproxReplace, Replace, 0, &Result, &ResultSize, &ResultIndex, 1);
	for (TYPEPTR *R=Result; R!=NULL && (*R)!=(TYPEPTR)~0UL; R++, R++){
	  if ((MaxWeight == (~0UL))||
	      ((*(R+1)) <= MaxWeight)){
	    if (out) {
	      fputs(Sep_Or, stdout);
	    }
	    out = 1;
	    PrintResults ((*R), 0, NULL, 0, Table);
	  }
	}
	if (!out) {
	  fputs(Sep_Uw, stdout);
	}
	fputs(Sep_Suff, stdout);
      } else {
	info = SearchStatic(LexedInitial, chaine);
	PrintResults(info, 0, myserver, 1, Table);
      }
      fflush(stdout);
    }
  }

  if (InputFile) {
    fclose (Input);
  }

  if (!LexedMemory) {
    fclose(Table);
  }
}

////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////
int main(int argn, char **argv) {
  char Prefix[LEXEDMAXSTRING];
  char Directory[LEXEDMAXSTRING];
  int i;

  strcpy (Directory, ".");
  strcpy (Prefix, "lexicon");

  Sep_Pref = (char *)calloc(LEXEDMAXSTRING+1, sizeof(char));
  Sep_Or   = (char *)calloc(LEXEDMAXSTRING+1, sizeof(char));
  Sep_Suff = (char *)calloc(LEXEDMAXSTRING+1, sizeof(char));
  Sep_Uw   = (char *)calloc(LEXEDMAXSTRING+1, sizeof(char));

  strcpy (Sep_Pref, "");
  strcpy (Sep_Or, "|");
  strcpy (Sep_Suff, "\n");
  strcpy (Sep_Uw, "<unknown>");
  strcpy (Delimiter, "	");

  LexedIndexation = 0;
  LexedMemory    = 1;
  Port[0]    = 0;
  nbreItem   = 0;

  if (argn > 1) {
    // generic options
    for (i = 1; argv[i]; i++) {
      if (argv[i][0] == '-') {
	if (!strcmp(argv[i]+1, "p")) {
	  if ((i+1 >= argn) || argv[i+1][0] == '-') {
	    Usage();
	  } else {
	    strcpy (Prefix, argv[++i]);
	  }
	} else if (!strcmp(argv[i]+1, "d")) {
	  if ((i+1 >= argn) || argv[i+1][0] == '-') {
	    Usage();
	  } else {
	    strcpy (Directory, argv[++i]);
	  }
	} else if (!strcmp(argv[i]+1, "v") || !strcmp(argv[i]+1, "-version")) {
	  fprintf(stderr, "%s version %s\n", PACKAGE, VERSION);
	  exit(0);
	} else if (!strcmp(argv[i]+1, "h") || !strcmp(argv[i]+1, "-help")) {
	  Usage();
	} else {
	  Usage();
	}
      } else {
	if (!strcmp(argv[i], "build")) {
	  return Build(argn, argv, i+1, Directory, Prefix);
	} else if (!strcmp(argv[i], "consult")) {
	  return Consult(argn, argv, i+1, Directory, Prefix);
	} else {
	  Usage();
	}
      }
    }
  } else {
    Usage();
  }
}
